from __future__ import annotations

from functools import partial, wraps

from pydiverse.transform._internal.errors import SubqueryError
from pydiverse.transform._internal.tree import verbs


class Pipeable:
    def __init__(self, f=None, calls=None):
        if f is not None:
            if calls is not None:
                raise ValueError
            self.calls = [f]
        else:
            self.calls = calls

    def __rshift__(self, other) -> Pipeable:
        """
        The pipe operator for chaining verbs.
        """

        if isinstance(other, Pipeable):
            return Pipeable(calls=self.calls + other.calls)
        elif callable(other):
            return Pipeable(calls=self.calls + [other])

        raise RuntimeError

    def __call__(self, arg):
        for c in self.calls:
            res = c(arg)
            if isinstance(res, Pipeable):
                res = res(arg)
            arg = res
        return arg


class inverse_partial(partial):
    """
    Just like partial, but the arguments get applied to the back instead of the front.
    This means that a function `def x(a, b, c)` decorated with `@inverse_partial(1, 2)`
    that gets called with `x(0)` is equivalent to calling `x(0, 1, 2)` on the non
    decorated function.
    """

    def __call__(self, /, *args, **keywords):
        keywords = {**self.keywords, **keywords}
        return self.func(*args, *self.args, **keywords)


def verb(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        return Pipeable(inverse_partial(fn, *args, **kwargs))

    return wrapper


def modify_ast(fn):
    @wraps(fn)
    def _fn(table, *args, **kwargs):
        new = fn(table, *args, **kwargs)
        assert new._ast != table._ast

        def _check_subquery(cache, ast_node):
            if cache.requires_subquery(new._ast):
                if not isinstance(ast_node, verbs.Alias):
                    raise SubqueryError(
                        f"Executing the `{new._ast.__class__.__name__.lower()}` verb "
                        f"on the table `{ast_node.name}` requires a subquery, which "
                        "is forbidden in transform by default.\n"
                        f"hint: Materialize the table `{ast_node.name}` before this "
                        "verb. If you are sure you want to do a subquery, put an "
                        "`>> alias()` before this verb. "
                    )
                return True
            return False

        # If a subquery is required, we put a marker in between
        assert new._ast.child == table._ast
        if _check_subquery(table._cache, table._ast):
            new._ast.child = verbs.SubqueryMarker(new._ast.child)
            new._cache = new._cache.update(new._ast.child)

        if isinstance(new._ast, verbs.Join):
            if _check_subquery(args[0]._cache, args[0]._ast):
                new._ast.right = verbs.SubqueryMarker(new._ast.right)
                right_cache = args[0]._cache.update(new._ast.right)
            else:
                right_cache = args[0]._cache

        new._cache = new._cache.update(
            new._ast,
            right_cache=right_cache if isinstance(new._ast, verbs.Join) else None,
        )

        return new

    return _fn



from __future__ import annotations

from pydiverse.transform._internal.ops.op import Operator
from pydiverse.transform._internal.ops.signature import Signature
from pydiverse.transform._internal.tree.types import (
    NUMERIC,
    Bool,
    Const,
    Decimal,
    Float,
    Int,
)

pow = Operator(
    "__pow__",
    Signature(Int(), Int(), return_type=Float()),
    Signature(Float(), Float(), return_type=Float()),
    Signature(Decimal(), Decimal(), return_type=Decimal()),
    doc="""
Computes the power x ** y.

Note
----
Polars throws on negative exponents in the integer case. A polars error like
`failed to convert X to u32` may be due to negative inputs to this function.
""",
)

neg = Operator(
    "__neg__",
    *(Signature(t, return_type=t) for t in NUMERIC),
    doc="The unary - (negation) operator (__neg__)",
)

pos = Operator(
    "__pos__",
    *(Signature(t, return_type=t) for t in NUMERIC),
    doc="The unary + operator (__pos__)",
)

abs = Operator(
    "abs",
    *(Signature(t, return_type=t) for t in NUMERIC),
    doc="Computes the absolute value.",
)

round = Operator(
    "round",
    *(Signature(t, Const(Int()), return_type=t) for t in NUMERIC),
    param_names=["self", "decimals"],
    default_values=[..., 0],
    doc="""
Rounds to a given number of decimals.

:param decimals:
    The number of decimals to round by.
""",
)

floor = Operator(
    "floor",
    Signature(Float(), return_type=Float()),
    Signature(Decimal(), return_type=Decimal()),
    doc="Returns the largest integer less than or equal to the input.",
)

ceil = Operator(
    "ceil",
    Signature(Float(), return_type=Float()),
    Signature(Decimal(), return_type=Decimal()),
    doc="Returns the smallest integer greater than or equal to the input.",
)

log = Operator(
    "log",
    Signature(Float(), return_type=Float()),
    doc="Computes the natural logarithm.",
)

exp = Operator(
    "exp",
    Signature(Float(), return_type=Float()),
    doc="Computes the exponential function.",
)

is_inf = Operator(
    "is_inf",
    Signature(Float(), return_type=Bool()),
    doc="""
Whether the number is infinite.

Note
----
This is currently only useful for backends supporting IEEE 754-floats. On
other backends it always returns False.
""",
)

is_not_inf = Operator("is_not_inf", Signature(Float(), return_type=Bool()))

is_nan = Operator("is_nan", Signature(Float(), return_type=Bool()))

is_not_nan = Operator("is_not_nan", Signature(Float(), return_type=Bool()))



from __future__ import annotations

import sqlalchemy as sqa
from sqlalchemy.dialects.postgresql import aggregate_order_by

from pydiverse.common import Bool, Dtype, Int64, String
from pydiverse.transform._internal.backend.sql import SqlImpl
from pydiverse.transform._internal.ops import ops
from pydiverse.transform._internal.tree import types
from pydiverse.transform._internal.tree.col_expr import Cast, ColFn


class PostgresImpl(SqlImpl):
    backend_name = "postgres"

    @classmethod
    def compile_cast(cls, cast: Cast, sqa_col: dict[str, sqa.Label]) -> Cast:
        compiled_val = cls.compile_col_expr(cast.val, sqa_col)

        if types.without_const(cast.val.dtype()).is_float():
            if cast.target_type.is_int():
                return sqa.func.trunc(compiled_val).cast(sqa.BigInteger())

            if cast.target_type == String():
                compiled = sqa.cast(compiled_val, sqa.String)
                return sqa.case(
                    (compiled == "NaN", "nan"),
                    (compiled == "Infinity", "inf"),
                    (compiled == "-Infinity", "-inf"),
                    else_=compiled,
                )

        if (
            types.without_const(cast.val.dtype()) == Bool()
            and cast.target_type == Int64()
        ):
            return sqa.cast(compiled_val, sqa.Integer)

        return sqa.cast(compiled_val, cls.sqa_type(cast.target_type))

    @classmethod
    def sqa_type(cls, pdt_type: Dtype):
        if isinstance(pdt_type, types.List):
            return sqa.types.ARRAY(item_type=cls.sqa_type())

        return super().sqa_type(pdt_type)

    @classmethod
    def fix_fn_types(
        cls, fn: ColFn, val: sqa.ColumnElement, *args: sqa.ColumnElement
    ) -> sqa.ColumnElement:
        if isinstance(fn.op, ops.DatetimeExtract | ops.DateExtract):
            return sqa.cast(val, sqa.BigInteger)
        elif fn.op == ops.sum:
            # postgres sometimes switches types for `sum`
            return sqa.cast(val, args[0].type)
        return val

    @classmethod
    def compile_ordered_aggregation(
        cls, *args: sqa.ColumnElement, order_by: list[sqa.UnaryExpression], impl
    ) -> sqa.ColumnElement:
        return impl(*args[:-1], aggregate_order_by(args[-1], *order_by))


with PostgresImpl.impl_store.impl_manager as impl:

    @impl(ops.less_than, String(), String())
    def _lt(x, y):
        return x < sqa.collate(y, "POSIX")

    @impl(ops.less_equal, String(), String())
    def _le(x, y):
        return x <= sqa.collate(y, "POSIX")

    @impl(ops.greater_than, String(), String())
    def _gt(x, y):
        return x > sqa.collate(y, "POSIX")

    @impl(ops.greater_equal, String(), String())
    def _ge(x, y):
        return x >= sqa.collate(y, "POSIX")

    @impl(ops.round)
    def _round(x, decimals=0):
        if decimals == 0:
            if isinstance(x.type, sqa.Integer):
                return x
            return sqa.func.ROUND(x, type_=x.type)

        if isinstance(x.type, sqa.Float):
            # Postgres doesn't support rounding of doubles to specific precision
            # -> Must first cast to numeric
            return sqa.func.ROUND(sqa.cast(x, sqa.Numeric), decimals, type_=sqa.Numeric)

        return sqa.func.ROUND(x, decimals, type_=x.type)

    @impl(ops.dt_second)
    def _dt_second(x):
        return sqa.func.FLOOR(sqa.extract("second", x), type_=sqa.Integer())

    @impl(ops.dt_millisecond)
    def _dt_millisecond(x):
        _1000 = sqa.literal_column("1000")
        return sqa.func.FLOOR(
            sqa.extract("milliseconds", x) % _1000, type_=sqa.Integer()
        )

    @impl(ops.horizontal_max, String(), String(), ...)
    def _horizontal_max(*x):
        return sqa.func.GREATEST(*(sqa.collate(e, "POSIX") for e in x))

    @impl(ops.horizontal_min, String(), String(), ...)
    def _least(*x):
        return sqa.func.LEAST(*(sqa.collate(e, "POSIX") for e in x))

    @impl(ops.any)
    @impl(ops.max, Bool())
    def _any(x):
        return sqa.func.BOOL_OR(x, type_=sqa.Boolean())

    @impl(ops.all)
    @impl(ops.min, Bool())
    def _all(x):
        return sqa.func.BOOL_AND(x, type_=sqa.Boolean())

    @impl(ops.is_nan)
    def _is_nan(x):
        return x == PostgresImpl.nan()

    @impl(ops.is_not_nan)
    def _is_not_nan(x):
        return x != PostgresImpl.nan()

    @impl(ops.dur_days)
    def _dur_days(x):
        sqa.func.extract("DAYS", x)

    @impl(ops.list_agg)
    def _list_agg(x):
        return sqa.func.array_agg(x)



from __future__ import annotations

from collections.abc import Generator, Iterable
from uuid import UUID


class AstNode:
    __slots__ = ["name"]

    name: str

    def clone(self) -> AstNode:
        return self._clone()[0]

    def _clone(self) -> tuple[AstNode, dict[AstNode, AstNode], dict[UUID, UUID]]: ...

    def iter_subtree_postorder(self) -> Iterable[AstNode]: ...

    # Iterates over the AST nodes in the subtree of the current node using preorder
    # traversal. By sending `True`, one can tell the DFS to not explore the subtree
    # any further.
    def iter_subtree_preorder(self) -> Generator[AstNode, bool | None, None]: ...



from __future__ import annotations

import functools

import polars as pl

from pydiverse.transform._internal.backend.targets import SqlAlchemy
from pydiverse.transform._internal.pipe.table import Table


def _cached_table(fn):
    cache = {}

    @functools.wraps(fn)
    def wrapped(df: pl.DataFrame, name: str):
        if name in cache:
            return cache[name]

        impl = fn(df, name)
        cache[name] = impl

        return impl

    return wrapped


@_cached_table
def polars_table(df: pl.DataFrame, name: str):
    return Table(df, name=name)


_sql_engine_cache = {}


def sql_table(df: pl.DataFrame, name: str, url: str, dtypes_map: dict | None = None):
    import sqlalchemy as sqa

    global _sql_engine_cache

    dtypes_map = dtypes_map or {}
    dtypes_map[pl.Float64] = sqa.Double

    if url in _sql_engine_cache:
        engine = _sql_engine_cache[url]
    else:
        engine = sqa.create_engine(url)
        _sql_engine_cache[url] = engine

    sql_dtypes = {}
    for col, dtype in zip(df.columns, df.dtypes, strict=True):
        if dtype in dtypes_map:
            sql_dtypes[col] = dtypes_map[dtype]

    df.write_database(
        name, engine, if_table_exists="replace", engine_options={"dtype": sql_dtypes}
    )
    return Table(name, SqlAlchemy(engine))


@_cached_table
def sqlite_table(df: pl.DataFrame, name: str):
    return sql_table(df, name, "sqlite:///:memory:")


@_cached_table
def duckdb_table(df: pl.DataFrame, name: str):
    return sql_table(df, name, "duckdb:///:memory:")


@_cached_table
def postgres_table(df: pl.DataFrame, name: str):
    url = "postgresql://sa:Pydiverse23@127.0.0.1:6543"

    return sql_table(df, name, url)


@_cached_table
def mssql_table(df: pl.DataFrame, name: str):
    from sqlalchemy.dialects.mssql import DATETIME2

    url = (
        "mssql+pyodbc://sa:PydiQuant27@127.0.0.1:1433"
        "/master?driver=ODBC+Driver+18+for+SQL+Server&encrypt=no"
    )
    return sql_table(df, name, url, dtypes_map={pl.Datetime(): DATETIME2()})


BACKEND_TABLES = {
    "polars": polars_table,
    "sqlite": sqlite_table,
    "duckdb": duckdb_table,
    "postgres": postgres_table,
    "mssql": mssql_table,
}



from __future__ import annotations

import inspect
from collections.abc import Callable, Iterable
from html import escape
from typing import Literal

import pandas as pd
import polars as pl
import sqlalchemy as sqa

from pydiverse.transform._internal import errors
from pydiverse.transform._internal.backend.table_impl import TableImpl
from pydiverse.transform._internal.backend.targets import Target
from pydiverse.transform._internal.errors import ColumnNotFoundError
from pydiverse.transform._internal.pipe.cache import Cache
from pydiverse.transform._internal.pipe.pipeable import Pipeable
from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import Col, ColName


class Table:
    __slots__ = ["_ast", "_cache"]

    """
    All attributes of a table are columns except for the `_ast` attribute
    which is a reference to the underlying abstract syntax tree.
    """

    def __init__(
        self,
        resource: pl.DataFrame | pl.LazyFrame | pd.DataFrame | sqa.Table | str | dict,
        backend: Target | None = None,
        *,
        name: str | None = None,
    ):
        """
        Creates a new table.

        :param resource:
            The data source to construct the table from. This can be a polars or pandas
            data frame, a python dictionary, a SQLAlchemy table or the name of a table
            in a SQL database.

        :param backend:
            The execution backend. This must be one of the pydiverse.transform backend
            objects, see :doc:`targets`. It may carry additional information how to
            interpret the *resource* argument, such as a SQLAlchemy engine.

        :param name:
            The name of the table. It is not required to give the table a name, but may
            make print output more readable.

        Examples
        --------
        **Python dictionary**.

        >>> t = pdt.Table(
        ...     {
        ...         "a": [4, 3, -35, 24, 105],
        ...         "b": [4, 4, 0, -23, 42],
        ...     },
        ...     name="T",
        ... )
        >>> t >> show()
        Table T, backend: PolarsImpl
        shape: (5, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 4   ┆ 4   │
        │ 3   ┆ 4   │
        │ -35 ┆ 0   │
        │ 24  ┆ -23 │
        │ 105 ┆ 42  │
        └─────┴─────┘

        **Polars data frame.**

        >>> df = pl.DataFrame(
        ...     {
        ...         "a": [4, 3, -35, 24, 105],
        ...         "b": ["a", "o", "---", "i23", "  "],
        ...     },
        ... )
        >>> t = pdt.Table(df, name="T")
        >>> t >> show()
        Table T, backend: PolarsImpl
        shape: (5, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 4   ┆ a   │
        │ 3   ┆ o   │
        │ -35 ┆ --- │
        │ 24  ┆ i23 │
        │ 105 ┆     │
        └─────┴─────┘

        **Pandas data frame.** Note that the data frame is converted to a polars data
        frame and the backend is polars.

        >>> import pandas as pd
        >>> df = pd.DataFrame(
        ...     {
        ...         "a": [4, 3, -35, 24, 105],
        ...         "b": ["a", "o", "---", "i23", "  "],
        ...     },
        ... )
        >>> t = pdt.Table(df, name="T")
        >>> t >> show()
        Table T, backend: PolarsImpl
        shape: (5, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 4   ┆ a   │
        │ 3   ┆ o   │
        │ -35 ┆ --- │
        │ 24  ┆ i23 │
        │ 105 ┆     │
        └─────┴─────┘

        **SQL.** Assuming you have a SQLAlchemy engine ``engine``, which is has a
        connection to a database containing a table ``t1`` in a schema ``s1``, you can
        create a pydiverse.transform Table from it as follows.

        >>> t = pdt.Table("t1", SqlAlchemy(engine, schema="s1"))
        >>> t >> show()
        Table t1, backend: PostgresImpl
        shape: (5, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 4   ┆ a   │
        │ 3   ┆ o   │
        │ -35 ┆ --- │
        │ 24  ┆ i23 │
        │ 105 ┆     │
        └─────┴─────┘

        Note that the name argument to the ``pdt.Table`` constructor was not specified,
        so transform used the name of the SQL table. This example of course assumes that
        a database connection is set up and the above table is already present in the
        database. For more information on how to set up a connection, see
        :doc:`/database_testing`.
        """

        if not isinstance(resource, TableImpl):
            errors.check_arg_type(
                pl.DataFrame | pl.LazyFrame | pd.DataFrame | sqa.Table | str | dict,
                "Table.__init__",
                "resource",
                resource,
            )
        errors.check_arg_type(Target | None, "Table.__init__", "backend", backend)
        errors.check_arg_type(str | None, "Table.__init__", "name", name)

        self._ast: AstNode = TableImpl.from_resource(resource, backend, name=name)
        self._cache = Cache.from_ast(self._ast)

    def __getitem__(self, key: str) -> Col:
        errors.check_arg_type(str, "Table.__getitem__", "key", key)
        return self.__getattr__(key)

    def __getattr__(self, name: str) -> Col:
        if name in ("__copy__", "__deepcopy__", "__setstate__", "__getstate__"):
            # for hasattr to work correctly on dunder methods
            raise AttributeError
        if name not in self._cache.name_to_uuid:
            raise ColumnNotFoundError(
                f"column `{name}` does not exist in table `{self._ast.name}`"
            )
        col = self._cache.cols[self._cache.name_to_uuid[name]]
        return Col(name, self._ast, col._uuid, col._dtype, col._ftype)

    def __setstate__(self, d):  # to avoid very annoying AttributeErrors
        for slot, val in d[1].items():
            setattr(self, slot, val)

    def __iter__(self) -> Iterable[Col]:
        cols = self._cache.selected_cols()
        yield from cols

    def __contains__(self, col: str | Col | ColName) -> bool:
        if isinstance(col, Col):
            return col._uuid in self._cache.uuid_to_name
        if isinstance(col, ColName):
            col = col.name
        return col in self._cache.name_to_uuid

    def __len__(self) -> int:
        return len(self._cache.name_to_uuid)

    def __rshift__(self, rhs) -> Table:
        """
        The pipe operator for chaining verbs.
        """

        if isinstance(rhs, Pipeable):
            return rhs(self)
        if isinstance(rhs, Callable):
            num_params = len(inspect.signature(rhs).parameters)
            if num_params != 1:
                raise TypeError(
                    "only functions with one parameter can be used in a pipe, got "
                    f"function with {num_params} parameters."
                )
            return self >> rhs(self)

        raise TypeError(
            f"found instance of invalid type `{type(rhs)}` in the pipe. \n"
            "hint: You can use a `Table` or a Callable taking a single argument in a "
            "pipe. If you use a Callable, it will receive the current table as an "
            "and must return a `Table`."
        )

    def __str__(self):
        from pydiverse.transform._internal.backend.targets import Polars
        from pydiverse.transform._internal.pipe.verbs import export, get_backend

        backend = get_backend(self._ast)
        try:
            df = self >> export(Polars(lazy=False))
        except Exception as e:
            return (
                f"Table {self._ast.name}, backend: {backend.__name__}\n"
                f"Failed to collect table.\n{type(e).__name__}: {str(e)}"
            )

        table_str = f"Table {self._ast.name}, backend: {backend.__name__}\n{df}"
        # TODO: cache the result for a polars backend

        return table_str

    def __dir__(self) -> list[str]:
        return [name for name in self._cache.name_to_uuid.keys()]

    def _repr_html_(self) -> str | None:
        html = (
            f"Table <code>{self._ast.name}</code> using"
            f" <code>{type(self._ast).__name__}</code> backend</br>"
        )
        try:
            from pydiverse.transform._internal.backend.targets import Polars
            from pydiverse.transform._internal.pipe.verbs import export

            # TODO: For lazy backend only show preview (eg. take first 20 rows)
            # TODO: also cache the table here for a polars backend. maybe we should call
            # collect() and manage this there?
            html += (self >> export(Polars()))._repr_html_()
        except Exception as e:
            html += (
                "</br><pre>Failed to collect table.\n"
                f"{escape(e.__class__.__name__)}: {escape(str(e))}</pre>"
            )
        return html

    def _repr_pretty_(self, p, cycle):
        p.text(str(self) if not cycle else "...")


def backend(table: Table) -> Literal["polars", "sqlite", "postgres", "duckdb", "mssql"]:
    return table._cache.backend


def is_sql_backed(table: Table) -> bool:
    return table._cache.backend != "polars"



from __future__ import annotations

import copy
import dataclasses
from typing import Literal
from uuid import UUID

from pydiverse.transform._internal.backend.table_impl import TableImpl
from pydiverse.transform._internal.ops.op import Ftype
from pydiverse.transform._internal.tree import types, verbs
from pydiverse.transform._internal.tree.ast import AstNode
from pydiverse.transform._internal.tree.col_expr import Col, ColFn


@dataclasses.dataclass(slots=True)
class Cache:
    name_to_uuid: dict[str, UUID]  # the selected columns, in order
    uuid_to_name: dict[UUID, str]  # again, only the selected columns + in order
    partition_by: list[UUID]
    derived_from: set[AstNode]  # for detecting invalid columns and self-joins
    cols: dict[UUID, Col]  # all columns in current scope (including hidden ones)

    # the following are only necessary for subquery detection
    limit: int
    group_by: set[UUID]
    is_filtered: bool

    backend: Literal["polars", "sqlite", "postgres", "duckdb", "mssql"]

    # For a column to be usable in an expression, the table it comes from must be an
    # ancestor of the current table AND the column's UUID must be in `all_cols` of the
    # current table (the latter is necessary because not every column is usable after
    # `summarize``)

    @staticmethod
    def from_ast(node: AstNode) -> Cache:
        if isinstance(node, verbs.Verb):
            if isinstance(node, verbs.Join):
                return Cache.from_ast(node.child).update(
                    node, right_cache=Cache.from_ast(node.right)
                )
            else:
                return Cache.from_ast(node.child).update(node)

        assert isinstance(node, TableImpl)
        return Cache(
            name_to_uuid={col.name: col._uuid for col in node.cols.values()},
            uuid_to_name={col._uuid: col.name for col in node.cols.values()},
            partition_by=[],
            derived_from={node},
            cols={col._uuid: col for col in node.cols.values()},
            limit=0,
            group_by=set(),
            is_filtered=False,
            backend=node.backend_name,
        )

    def update(self, node: verbs.Verb, *, right_cache: Cache | None = None) -> Cache:
        """
        Returns a new cache for `node`, assuming `self` is the cache of `node.child`.
        Does not modify `self`.
        """

        res = copy.copy(self)

        if isinstance(node, verbs.Alias):
            if node.uuid_map is not None:
                res.name_to_uuid = {
                    name: node.uuid_map[uid] for name, uid in self.name_to_uuid.items()
                }
                res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}
                res.cols = {
                    node.uuid_map[uid]: Col(
                        col.name, node, node.uuid_map[uid], col._dtype, col._ftype
                    )
                    for uid, col in self.cols.items()
                }
                res.partition_by = [node.uuid_map[uid] for uid in self.partition_by]
                res.derived_from = set()

        elif isinstance(node, verbs.Select):
            selected_uuids = set(col._uuid for col in node.select)
            res.uuid_to_name = {
                uid: name
                for uid, name in self.uuid_to_name.items()
                if uid in selected_uuids
            }
            res.name_to_uuid = {name: uid for uid, name in res.uuid_to_name.items()}

        elif isinstance(node, verbs.Rename):
            res.name_to_uuid = {
                (new_name if (new_name := node.name_map.get(name)) else name): uid
                for name, uid in self.name_to_uuid.items()
            }
            res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}

        elif isinstance(node, verbs.Mutate):
            res.cols = self.cols | {
                uid: Col(name, node, uid, val.dtype(), val.ftype(agg_is_window=True))
                for name, val, uid in zip(
                    node.names, node.values, node.uuids, strict=True
                )
            }
            res.name_to_uuid = self.name_to_uuid | {
                name: uid for name, uid in zip(node.names, node.uuids, strict=True)
            }
            res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}

        elif isinstance(node, verbs.Filter):
            res.is_filtered = True

        elif isinstance(node, verbs.GroupBy):
            res.partition_by = [col._uuid for col in node.group_by]
            if node.add:
                res.partition_by = self.partition_by + res.partition_by

        elif isinstance(node, verbs.Ungroup):
            res.partition_by = []

        elif isinstance(node, verbs.Summarize):
            overwritten = {
                col_name for col_name in node.names if col_name in self.name_to_uuid
            }
            cols = {
                self.uuid_to_name[uid]: self.cols[uid]
                for uid in self.partition_by
                if self.uuid_to_name[uid] not in overwritten
            } | {
                name: Col(name, node, uid, val.dtype(), val.ftype())
                for name, val, uid in zip(
                    node.names, node.values, node.uuids, strict=True
                )
            }

            res.cols = {col._uuid: col for _, col in cols.items()}
            res.name_to_uuid = {name: col._uuid for name, col in cols.items()}
            res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}
            res.group_by = res.group_by | set(res.partition_by)
            res.partition_by = []

        elif isinstance(node, verbs.SliceHead):
            res.limit = node.n

        elif isinstance(node, verbs.Join):
            assert right_cache is not None

            res.cols = self.cols | right_cache.cols
            res.name_to_uuid = self.name_to_uuid | {
                name + node.suffix: uid
                for name, uid in right_cache.name_to_uuid.items()
            }
            res.uuid_to_name = {uid: name for name, uid in res.name_to_uuid.items()}

            res.derived_from = self.derived_from | right_cache.derived_from
            res.limit = 0
            res.group_by = set()

        elif isinstance(node, verbs.SubqueryMarker):
            res.cols = {
                uid: Col(
                    col.name,
                    node.child,
                    uid,
                    types.without_const(col._dtype),
                    Ftype.ELEMENT_WISE,
                )
                for uid, col in self.cols.items()
            }
            res.limit = 0
            res.group_by = set()
            res.is_filtered = False

        assert len(res.name_to_uuid) == len(res.uuid_to_name)
        res.derived_from = res.derived_from | {node}

        return res

    # Returns whether applying `node` to this table would require it to be wrapped in a
    # subquery. (so `self` is the old cache and `node` the new AST)
    def requires_subquery(self, node: verbs.Verb) -> bool:
        if self.backend == "polars":
            return False

        return (
            (
                isinstance(
                    node,
                    verbs.Filter
                    | verbs.Summarize
                    | verbs.Arrange
                    | verbs.GroupBy
                    | verbs.Join,
                )
                and self.limit != 0
            )
            or (
                isinstance(node, verbs.Mutate)
                and any(
                    any(
                        col.ftype(agg_is_window=True) in (Ftype.WINDOW, Ftype.AGGREGATE)
                        for col in fn.iter_subtree()
                        if isinstance(col, Col)
                    )
                    for fn in node.iter_col_nodes()
                    if (
                        isinstance(fn, ColFn)
                        and fn.op.ftype in (Ftype.AGGREGATE, Ftype.WINDOW)
                    )
                )
            )
            or (
                isinstance(node, verbs.Filter)
                and any(
                    col.ftype(agg_is_window=True) == Ftype.WINDOW
                    for col in node.iter_col_nodes()
                    if isinstance(col, Col)
                )
            )
            or (
                isinstance(node, verbs.Summarize)
                and (
                    (self.group_by and self.group_by != set(self.partition_by))
                    or any(
                        (
                            col.ftype(agg_is_window=False)
                            in (Ftype.WINDOW, Ftype.AGGREGATE)
                        )
                        for col in node.iter_col_nodes()
                        if isinstance(col, Col)
                    )
                    or any(
                        self.cols[uid].ftype() == Ftype.WINDOW
                        for uid in self.partition_by
                    )
                )
            )
            or (
                isinstance(node, verbs.Join)
                and (
                    self.group_by
                    or (
                        (
                            node.how == "full"
                            or (node.right in self.derived_from and node.how == "left")
                        )
                        and any(
                            types.is_const(self.cols[uid].dtype())
                            for uid in self.uuid_to_name.keys()
                        )
                    )
                    or any(
                        self.cols[uid].ftype() == Ftype.WINDOW
                        for uid in self.uuid_to_name.keys()
                    )
                    or any(
                        col.ftype() != Ftype.ELEMENT_WISE and col._uuid in self.cols
                        for col in node.on.iter_subtree()
                        if isinstance(col, Col)
                    )
                    or (self.is_filtered and node.how == "full")
                )
            )
        )

    def selected_cols(self) -> list[Col]:
        return [self.cols[uid] for uid in self.uuid_to_name.keys()]



from __future__ import annotations

import pydiverse.transform as pdt
from pydiverse.transform._internal.pipe.c import C
from pydiverse.transform._internal.pipe.verbs import mutate, select
from tests.fixtures.backend import skip_backends
from tests.util.assertion import assert_result_equal


def test_exp(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(**{c.name: pdt.max(-700, pdt.min(700, c)).exp() for c in t}),
    )


def test_log(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: pdt.max(1e-16, c).log() for c in t}),
    )


def test_abs(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.abs() for c in t}),
    )


def test_round(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.round() for c in t}),
    )


def test_add(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(**{f"add_{c.name}_{d.name}": c + d for d in t for c in t}),
    )


def test_sub(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(**{f"sub_{c.name}_{d.name}": c - d for d in t for c in t}),
    )


def test_neg(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{f"neg_{c.name}": -c for c in t}),
    )


def test_mul(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(**{f"mul_{c.name}_{d.name}": c * d for d in t for c in t}),
    )


def test_div(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(
            **{c.name: pdt.when(c.abs() < 1e-50).then(1e-50).otherwise(c) for c in t}
        )
        >> (
            lambda s: mutate(**{f"div_{c.name}_{d.name}": c / d for d in s for c in s})
        ),
    )


def test_decimal(df_num):
    # TODO: test the decimal here
    assert_result_equal(df_num, lambda t: t >> mutate(u=t.f + t.g, z=t.f * t.g))


def test_floor(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.floor() for c in t}),
    )


def test_ceil(df_num):
    assert_result_equal(
        df_num,
        lambda t: t >> mutate(**{c.name: c.ceil() for c in t}),
    )


@skip_backends("mssql")
def test_inf_lit(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> select()
        >> mutate(
            inf=float("inf"),
            neg_inf=float("-inf"),
        )
        >> mutate(
            inf_str=C.inf.cast(pdt.String()),
            neg_inf_str=C.neg_inf.cast(pdt.String()),
        )
        >> mutate(
            inf_back=C.inf_str.cast(pdt.Float64()),
            neg_inf_back=C.neg_inf_str.cast(pdt.Float64()),
        ),
    )


@skip_backends("mssql", "sqlite")
def test_nan_lit(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> select()
        >> mutate(nan=float("nan"))
        >> mutate(nan_str=C.nan.cast(pdt.String()))
        >> mutate(nan_back=C.nan_str.cast(pdt.Float64())),
    )


@skip_backends("mssql")
def test_is_inf(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(inf=float("inf"))
        >> mutate(
            inf_is_inf=C.inf.is_inf(),
            **{c.name + "is_inf": c.is_inf() for c in t},
            inf_is_not_inf=C.inf.is_not_inf(),
            **{c.name + "is_not_inf": c.is_not_inf() for c in t},
        ),
    )


@skip_backends("mssql", "sqlite")
def test_is_nan(df_num):
    assert_result_equal(
        df_num,
        lambda t: t
        >> mutate(nan=float("nan"))
        >> mutate(
            nan_is_nan=C.nan.is_nan(),
            **{c.name + "is_nan": c.is_nan() for c in t},
            nan_is_not_nan=C.nan.is_not_nan(),
            **{c.name + "is_not_nan": c.is_not_nan() for c in t},
        ),
    )


def test_int_pow(df_int):
    assert_result_equal(
        df_int, lambda t: t >> mutate(u=pdt.min(t.a, 10) ** pdt.min(t.b.abs(), 5))
    )
